"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const chai = require("chai");
const sinon = require("sinon");
const varsnap = require("../varsnap/core");
const configTools = require("./config_tools");
const expect = chai.expect;
function example(x, y) {
    return x * y;
}
describe('Util', function () {
    describe('getConfig', function () {
        it('can read from env vars', function () {
            const config = varsnap.Util.getConfig(configTools.producerConfig(), 'env');
            expect(config).to.equal('production');
        });
        it('will lowercase configs', function () {
            const configs = {
                'varsnap': 'true',
                'env': 'TEST',
            };
            const config = varsnap.Util.getConfig(configs, 'env');
            expect(config).to.equal('test');
        });
        it('will work with non-string configs', function () {
            const configs = {
                'varsnap': true,
                'env': 'test',
            };
            const config = varsnap.Util.getConfig(configs, 'varsnap');
            expect(config).to.equal('true');
        });
        it('will default to empty string if there are no configs', function () {
            const config = varsnap.Util.getConfig(undefined, 'env'); // eslint-disable-line @typescript-eslint/no-explicit-any
            expect(config).to.equal('');
        });
        it('will default to empty string if config is not defined', function () {
            const configs = {
                'varsnap': true,
                'env': 'test',
            };
            const config = varsnap.Util.getConfig(configs, 'producerToken');
            expect(config).to.equal('');
        });
    });
    describe('getSignature', function () {
        it('returns a valid signature', function () {
            const signature = varsnap.Util.getSignature(varsnap.Util.getConfig);
            expect(signature).to.include('javascript');
            expect(signature).to.include(varsnap.version);
            expect(signature).to.include('getConfig');
        });
    });
    describe('limitString', function () {
        it('returns a short string', function () {
            const x = 'a';
            expect(varsnap.Util.limitString(x)).to.equal(x);
            const y = 'a'.repeat(30);
            expect(varsnap.Util.limitString(y)).to.equal(y);
        });
        it('returns a shortened string', function () {
            const x = 'a'.repeat(40);
            const limited = varsnap.Util.limitString(x);
            expect(limited.length).to.equal(30);
            expect(limited.substring(0, 27)).to.equal('a'.repeat(27));
            expect(limited.substring(27, 30)).to.equal('...');
        });
    });
});
describe('Producer', function () {
    beforeEach(function () {
        this.producer = new varsnap.Producer(example, configTools.producerConfig());
    });
    describe('isEnabled', function () {
        it('will return false if varsnap is not enabled', function () {
            this.producer.config.varsnap = 'false';
            expect(this.producer.isEnabled()).to.be.false;
        });
        it('will return false if env is not production', function () {
            this.producer.config.env = 'development';
            expect(this.producer.isEnabled()).to.be.false;
        });
        it('will return false if producer token is not set', function () {
            this.producer.config.producerToken = '';
            expect(this.producer.isEnabled()).to.be.false;
        });
        it('will return true if all configs are correct', function () {
            expect(this.producer.isEnabled()).to.be.true;
        });
    });
    describe('serialize', function () {
        it('can serialize data', function () {
            expect(varsnap.Producer.serialize(5)).to.equal('5');
            expect(varsnap.Producer.serialize({ '1': [2] })).to.equal('{"1":[2]}');
        });
    });
    describe('produce', function () {
        beforeEach(function () {
            configTools.setProducerConfig();
            this.ajax = sinon.stub(varsnap.Util, 'ajax');
        });
        afterEach(function () {
            this.ajax.restore();
        });
        it('will send data', function () {
            this.ajax.resolves('asdf');
            this.producer.produce([2, 3], 6);
            sinon.assert.called(this.ajax);
        });
    });
});
describe('Consumer', function () {
    beforeEach(function () {
        this.consumer = new varsnap.Consumer(example, configTools.consumerConfig());
    });
    describe('isEnabled', function () {
        it('will return false if varsnap is not enabled', function () {
            this.consumer.config.varsnap = 'false';
            expect(this.consumer.isEnabled()).to.be.false;
        });
        it('will return false if env is not development', function () {
            this.consumer.config.env = 'production';
            expect(this.consumer.isEnabled()).to.be.false;
        });
        it('will return false if consumer token is not set', function () {
            this.consumer.config.consumerToken = '';
            expect(this.consumer.isEnabled()).to.be.false;
        });
        it('will return true if all configs are correct', function () {
            expect(this.consumer.isEnabled()).to.be.true;
        });
    });
    describe('deserialize', function () {
        it('can deserialize data', function () {
            expect(varsnap.Consumer.deserialize('5')).to.equal(5);
            expect(varsnap.Consumer.deserialize('{"1":[2]}')).to.deep.equal({ '1': [2] });
        });
    });
    describe('consumeOnce', function () {
        beforeEach(function () {
            configTools.setConsumerConfig();
            this.ajax = sinon.stub(varsnap.Util, 'ajax');
        });
        afterEach(function () {
            this.ajax.restore();
        });
        it('will not run if the consumer is not enabled', function () {
            this.consumer.config.env = 'production';
            this.consumer.consumeOnce();
            sinon.assert.callCount(this.ajax, 0);
        });
        it('will make an ajax request and report it', function () {
            const data = {
                'id': 'abcd',
                'inputs': '[2, 3]',
                'prod_outputs': '4',
            };
            this.ajax.resolves(data);
            this.consumer.consumeOnce();
            sinon.assert.called(this.ajax);
        });
    });
});
describe('varsnap', function () {
    beforeEach(function () {
        configTools.setProducerConfig();
        varsnap.Util.resetConsumers();
        varsnap.Util.resetProducers();
    });
    it('creates a consumer and producer', function () {
        const wrapped = varsnap.varsnap(example);
        expect(wrapped.producer.targetFunc).to.equal(example);
        expect(wrapped.producer.config).to.equal(varsnap.varsnap.config);
        expect(wrapped.consumer.targetFunc).to.equal(example);
        expect(wrapped.consumer.config).to.equal(varsnap.varsnap.config);
    });
    it('wraps a function', function () {
        const wrapped = varsnap.varsnap(example);
        expect(wrapped(2, 4)).to.equal(8);
    });
    it('saves consumers and producers', function () {
        expect(varsnap.Util.getConsumers().length).to.equal(0);
        expect(varsnap.Util.getConsumers().length).to.equal(0);
        varsnap.varsnap(example);
        expect(varsnap.Util.getConsumers().length).to.equal(1);
        expect(varsnap.Util.getConsumers().length).to.equal(1);
    });
});
describe('version', function () {
    it('has a version', function () {
        expect(varsnap.version).to.be;
    });
});
