import deepEqual = require('fast-deep-equal');
import querystring = require('querystring');
import xhr = require('request');

const produceURL = 'https://www.varsnap.com/api/snap/produce/';
const consumeURL = 'https://www.varsnap.com/api/snap/consume/';
const produceTrialURL = 'https://www.varsnap.com/api/trial/produce/';
export const version = 'v1.3.3';

export interface Config {
  varsnap: string|boolean;
  env: string;
  producerToken?: string;
  consumerToken?: string;
  logger?: Logger;
}
type ConfigKeys = keyof Config;
const defaultConfig: Config = {
  varsnap: 'true',
  env: 'development',
  producerToken: '',
  consumerToken: '',
  logger: console, // eslint-disable-line no-console
};

let Producers: Producer[] = [];
let Consumers: Consumer[] = [];

export interface Logger {
  debug(s: string): void;
  log(s: string): void;
  error(s: string): void;
}

export class Util {
  public static logger: Logger = console; // eslint-disable-line no-console

  static getConfig(config: Config, configKey: ConfigKeys): string {
    if (!config) {
      return '';
    }
    let val = config[configKey];
    if (!val) {
      return '';
    }
    val = val.toString().toLowerCase();
    return val;
  }

  static getSignature(targetFunc: TargetFunction): string {
    let signature = targetFunc.toString().split('{')[0];
    signature = signature.replace(/^\s+|\s+$/g, '');
    return 'javascript.' + version + '.' + signature;
  }

  static ajax(url: string, requestData: object): Promise<object> {
    const options = {
      method: 'POST',
      uri: url,
      body: querystring.stringify(requestData as querystring.ParsedUrlQueryInput),
      headers: {'Content-Type': 'application/x-www-form-urlencoded'},
    };
    const ajaxPromise = new Promise<object>((resolve, reject) => {
      xhr(options, (err, resp, body) => {
        if (err) {
          return reject(err);
        }
        const responseCode = Math.round(resp.statusCode / 100);
        if (responseCode != 2) {
          return reject(new Error('Status: ' + resp.statusCode));
        }
        body = JSON.parse(body);
        return resolve(body);
      });
    });
    return ajaxPromise;
  }

  static debug(data: string): void {
    Util.logger.debug(data);
  }

  static log(data: string): void {
    Util.logger.log(data);
  }

  static error(data: string): void {
    Util.logger.error(data);
  }

  static resetProducers(): void {
    Producers = [];
  }

  static resetConsumers(): void {
    Consumers = [];
  }

  static getProducers(): Producer[] {
    return Producers;
  }

  static getConsumers(): Consumer[] {
    return Consumers;
  }

  static limitString(x: TargetFunctionInput[] | TargetFunctionOutput): string {
    const limitedString = String(x);
    const limit = 30;
    const ellipsis = '...';
    if (limitedString.length <= limit) {
      return limitedString;
    }
    return limitedString.substring(0, 30 - ellipsis.length) + ellipsis;
  }
}

export class Producer {
  public targetFunc: TargetFunction;
  public config: Config;

  constructor(targetFunc: TargetFunction, config: Config) {
    this.targetFunc = targetFunc;
    this.config = config;
  }

  isEnabled(): boolean {
    if (Util.getConfig(this.config, 'varsnap') !== 'true') {
      return false;
    }
    if (Util.getConfig(this.config, 'env') !== 'production') {
      return false;
    }
    if (!Util.getConfig(this.config, 'producerToken')) {
      return false;
    }
    return true;
  }

  static serialize(data: TargetFunctionInput[] | TargetFunctionOutput): string {
    return JSON.stringify(data);
  }

  produce(args: TargetFunctionInput[], output: TargetFunctionOutput): void {
    if (!this.isEnabled()) {
      return;
    }
    Util.debug('Sending call to Varsnap');
    const data = {
      'producer_token': Util.getConfig(this.config, 'producerToken'),
      'signature': Util.getSignature(this.targetFunc),
      'inputs': Producer.serialize(args),
      'prod_outputs': Producer.serialize(output),
    };
    Util.ajax(produceURL, data).catch((err) => {
      if(err.statusCode === 404) {
        Util.error('Cannot produce to Varsnap; check your producer token');
      } else {
        Util.error(err.error);
      }
    });
  }
}

interface ConsumeSnapResponse {
  status: string;
  results: Snap[];
}

interface Snap {
  id: string;
  inputs: string;
  prod_outputs: string;
}

interface ProduceTrialResponse {
  status: string;
  trial_url: string;
}

export class Consumer {
  public targetFunc: TargetFunction;
  public config: Config;

  constructor(targetFunc: TargetFunction, config: Config) {
    this.targetFunc = targetFunc;
    this.config = config;
  }

  isEnabled(): boolean {
    if (Util.getConfig(this.config, 'varsnap') !== 'true') {
      return false;
    }
    if (Util.getConfig(this.config, 'env') !== 'development') {
      return false;
    }
    if (!Util.getConfig(this.config, 'consumerToken')) {
      return false;
    }
    return true;
  }

  static deserialize(data: string): TargetFunctionInput[] | TargetFunctionOutput | undefined {
    if (data === '') {
      return undefined;
    }
    return JSON.parse(data);
  }

  consumeOnce(): Promise<boolean> {
    if (!this.isEnabled()) {
      const errorPromise = new Promise<boolean>((resolve) => {
        return resolve(false);
      });
      return errorPromise;
    }
    const data = {
      'consumer_token': Util.getConfig(this.config, 'consumerToken'),
      'signature': Util.getSignature(this.targetFunc),
    };
    const ajax = Util.ajax(consumeURL, data);
    const consume = ajax.then((rawResponse: object): boolean => {
      const response = rawResponse as ConsumeSnapResponse;
      if (!response || response.status !== 'ok') {
        return false;
      }
      let allMatches = true;
      for(const snap of response.results) {
        const inputs: TargetFunctionInput[] = Consumer.deserialize(snap['inputs']) as TargetFunctionInput[];
        const prodOutputs: TargetFunctionOutput = Consumer.deserialize(snap['prod_outputs']) as TargetFunctionOutput;
        let localOutputs: TargetFunctionOutput;
        try {
          localOutputs = this.targetFunc(...inputs);
        }
        catch(err) {
          localOutputs = err;
        }
        const matches = deepEqual(prodOutputs, localOutputs);
        this.reportCentral(snap, localOutputs, matches).then((trialURL) => {
          this.reportLog(snap['id'], inputs, prodOutputs, localOutputs, matches, trialURL);
        });
        allMatches = allMatches && matches;
      }
      return allMatches;
    }).catch((error: object) => {
      Util.error('Cannot contact Varsnap: ' + error);
      return false;
    });
    return consume;
  }

  reportCentral(snap: Snap, localOutputs: TargetFunctionOutput, matches: boolean): Promise<string|boolean> {
    const data = {
      'consumer_token': Util.getConfig(this.config, 'consumerToken'),
      'snap_id': snap.id,
      'test_outputs': Producer.serialize(localOutputs),
      'matches': matches,
    };
    const ajax = Util.ajax(produceTrialURL, data);
    const report = ajax.then((rawResponse: object): string|boolean => {
      const response = rawResponse as ProduceTrialResponse;
      if (!response || response.status !== 'ok') {
        return false;
      }
      const trialURL = response['trial_url'];
      return trialURL;
    }).catch((error: object) => {
      Util.error('Cannot upload test results to Varsnap: ' + error);
      return false;
    });
    return report;
  }

  reportLog(snapID: string, inputs: TargetFunctionInput[], prodOutputs: TargetFunctionOutput, localOutputs: TargetFunctionOutput, matches: boolean, trialURL: string|boolean): void {
    const functionName = Util.getSignature(this.targetFunc);
    Util.log('Testing with Varsnap uuid:     ' + snapID);
    Util.log('Function:                      ' + functionName);
    Util.log('Function input args:           ' + Util.limitString(inputs));
    Util.log('Production function outputs:   ' + Util.limitString(prodOutputs));
    Util.log('Your function outputs:         ' + Util.limitString(localOutputs));
    Util.log('Matching outputs:              ' + matches);
    let trialURLView = 'Cannot upload trial';
    if(trialURL) {
      trialURLView = String(trialURL);
    }
    Util.log('Trial URL:                     ' + trialURLView);
    Util.log('');
  }
}

export type TargetFunctionInput = any; // eslint-disable-line @typescript-eslint/no-explicit-any
export type TargetFunctionOutput = any; // eslint-disable-line @typescript-eslint/no-explicit-any
export interface TargetFunction {
  (...args: TargetFunctionInput[]): TargetFunctionOutput;
}
export interface WrappedFunction {
  (...args: TargetFunctionInput[]): TargetFunctionOutput;
  producer: Producer;
  consumer: Consumer;
}

export function varsnap(func: TargetFunction): TargetFunction|WrappedFunction {
  if (varsnap.config === undefined) {
    console.error('Varsnap has no config; disabling.');
    return func;
  }
  if (varsnap.config.logger) {
    Util.logger = varsnap.config.logger;
  }
  const producer = new Producer(func, varsnap.config);
  const consumer = new Consumer(func, varsnap.config);

  const wrapped = function(...args: TargetFunctionInput[]): TargetFunctionOutput {
    let result: TargetFunctionOutput;
    try {
      result = func(...args);
    }
    catch(err) {
      producer.produce(args, err);
      throw(err);
    }
    producer.produce(args, result);
    return result;
  };

  wrapped.producer = producer;
  wrapped.consumer = consumer;

  Producers.push(producer);
  Consumers.push(consumer);
  return wrapped;
}
varsnap.config = defaultConfig;
varsnap.version = version;

export async function runTests(): Promise<boolean> {
  const consumers = Util.getConsumers();
  let status = true;
  for(const consumer of consumers) {
    const testStatus = await consumer.consumeOnce();
    status = status && testStatus;
  }
  return status;
}
varsnap.runTests = runTests;
